/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */
#include <stdlib.h>
#include <bsd/string.h>
#include <unistd.h>

#include <doca_argp.h>
#include <doca_log.h>

#include <utils.h>

#include "app_shield_agent_core.h"

DOCA_LOG_REGISTER(APSH_APP::Core);

/* This value is guaranteed to be 253 on Linux, and 16 bytes on Windows */
#define MAX_HOSTNAME_LEN 253

static doca_error_t
pid_callback(void *config, void *param)
{
	struct apsh_config *conf = (struct apsh_config *)config;

	conf->pid = (unsigned int)*(int *)param;
	return DOCA_SUCCESS;
}

static doca_error_t
hash_map_callback(void *config, void *param)
{
	struct apsh_config *conf = (struct apsh_config *)config;
	size_t size = sizeof(conf->exec_hash_map_path);

	if (strnlen(param, size) >= size) {
		DOCA_LOG_ERR("Execute hash map argument too long. Must be <=%zu long.", size - 1);
		return DOCA_ERROR_INVALID_VALUE;
	}
	strcpy(conf->exec_hash_map_path, param);

	if (access(conf->exec_hash_map_path, F_OK) == -1) {
		DOCA_LOG_ERR("Execute hash map json file not found %s", conf->exec_hash_map_path);
		return DOCA_ERROR_NOT_FOUND;
	}
	return DOCA_SUCCESS;
}

static doca_error_t
memr_callback(void *config, void *param)
{
	struct apsh_config *conf = (struct apsh_config *)config;
	size_t size = sizeof(conf->system_mem_region_path);

	if (strnlen(param, size) >= size) {
		DOCA_LOG_ERR("System memory regions map argument too long. Must be <=%zu long.", size - 1);
		return DOCA_ERROR_INVALID_VALUE;
	}
	strcpy(conf->system_mem_region_path, param);

	if (access(conf->system_mem_region_path, F_OK) == -1) {
		DOCA_LOG_ERR("System memory regions map json file not found %s", conf->system_mem_region_path);
		return DOCA_ERROR_NOT_FOUND;
	}
	return DOCA_SUCCESS;
}

static doca_error_t
vuid_callback(void *config, void *param)
{
	struct apsh_config *conf = (struct apsh_config *)config;
	size_t size = sizeof(conf->system_vuid);

	if (strnlen(param, size) >= size) {
		DOCA_LOG_ERR("system VUID argument too long. Must be <=%zu long.", size - 1);
		return DOCA_ERROR_INVALID_VALUE;
	}
	strcpy(conf->system_vuid, param);
	return DOCA_SUCCESS;
}

static doca_error_t
dma_callback(void *config, void *param)
{
	struct apsh_config *conf = (struct apsh_config *)config;
	size_t size = sizeof(conf->dma_dev_name);

	if (strnlen(param, size) >= size) {
		DOCA_LOG_ERR("DMA device name argument too long. Must be <=%zu long.", size - 1);
		return DOCA_ERROR_INVALID_VALUE;
	}
	strcpy(conf->dma_dev_name, param);
	return DOCA_SUCCESS;
}

static doca_error_t
os_syms_callback(void *config, void *param)
{
	struct apsh_config *conf = (struct apsh_config *)config;
	size_t size = sizeof(conf->system_os_symbol_map_path);

	if (strnlen(param, size) >= size) {
		DOCA_LOG_ERR("System os symbols map argument too long. Must be <=%zu long.", size - 1);
		return DOCA_ERROR_INVALID_VALUE;
	}
	strcpy(conf->system_os_symbol_map_path, param);

	if (access(conf->system_os_symbol_map_path, F_OK) == -1) {
		DOCA_LOG_ERR("System os symbols map json file not found %s", conf->system_os_symbol_map_path);
		return DOCA_ERROR_NOT_FOUND;
	}
	return DOCA_SUCCESS;
}

static doca_error_t
os_type_callback(void *config, void *param)
{
	struct apsh_config *conf = (struct apsh_config *)config;
	char *str_param = (char *)param;

	if (!strcasecmp(str_param, "windows"))
		conf->os_type = DOCA_APSH_SYSTEM_WINDOWS;
	else if (!strcasecmp(str_param, "linux"))
		conf->os_type = DOCA_APSH_SYSTEM_LINUX;
	else {
		DOCA_LOG_ERR("OS type is not windows/linux (case insensitive)");
		return DOCA_ERROR_NOT_SUPPORTED;
	}
	return DOCA_SUCCESS;
}

static doca_error_t
time_callback(void *config, void *param)
{
	struct apsh_config *conf = (struct apsh_config *)config;

	conf->time_interval = *(int *)param;
	return DOCA_SUCCESS;
}

void
register_apsh_params()
{
	doca_error_t ret;
	struct doca_argp_param *pid_param, *hash_map_param, *memr_param, *vuid_param, *dma_param, *os_syms_param;
	struct doca_argp_param *time_param, *os_type_param;

	/* Create and register pid param */
	ret = doca_argp_param_create(&pid_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to create ARGP param: %s", doca_get_error_string(ret));
	doca_argp_param_set_short_name(pid_param, "p");
	doca_argp_param_set_long_name(pid_param, "pid");
	doca_argp_param_set_description(pid_param, "Process ID of process to be attested");
	doca_argp_param_set_callback(pid_param, pid_callback);
	doca_argp_param_set_type(pid_param, DOCA_ARGP_TYPE_INT);
	doca_argp_param_set_mandatory(pid_param);
	ret = doca_argp_register_param(pid_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to register program param: %s", doca_get_error_string(ret));

	/* Create and register regex pci address param */
	ret = doca_argp_param_create(&hash_map_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to create ARGP param: %s", doca_get_error_string(ret));
	doca_argp_param_set_short_name(hash_map_param, "e");
	doca_argp_param_set_long_name(hash_map_param, "ehm");
	doca_argp_param_set_arguments(hash_map_param, "<path>");
	doca_argp_param_set_description(hash_map_param, "Exec hash map path");
	doca_argp_param_set_callback(hash_map_param, hash_map_callback);
	doca_argp_param_set_type(hash_map_param, DOCA_ARGP_TYPE_STRING);
	doca_argp_param_set_mandatory(hash_map_param);
	ret = doca_argp_register_param(hash_map_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to register program param: %s", doca_get_error_string(ret));

	/* Create and register system memory map param */
	ret = doca_argp_param_create(&memr_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to create ARGP param: %s", doca_get_error_string(ret));
	doca_argp_param_set_short_name(memr_param, "m");
	doca_argp_param_set_long_name(memr_param, "memr");
	doca_argp_param_set_arguments(memr_param, "<path>");
	doca_argp_param_set_description(memr_param, "System memory regions map");
	doca_argp_param_set_callback(memr_param, memr_callback);
	doca_argp_param_set_type(memr_param, DOCA_ARGP_TYPE_STRING);
	doca_argp_param_set_mandatory(memr_param);
	ret = doca_argp_register_param(memr_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to register program param: %s", doca_get_error_string(ret));

	/* Create and register VUID param */
	ret = doca_argp_param_create(&vuid_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to create ARGP param: %s", doca_get_error_string(ret));
	doca_argp_param_set_short_name(vuid_param, "f");
	doca_argp_param_set_long_name(vuid_param, "vuid");
	doca_argp_param_set_description(vuid_param, "VUID of the System device");
	doca_argp_param_set_callback(vuid_param, vuid_callback);
	doca_argp_param_set_type(vuid_param, DOCA_ARGP_TYPE_STRING);
	doca_argp_param_set_mandatory(vuid_param);
	ret = doca_argp_register_param(vuid_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to register program param: %s", doca_get_error_string(ret));

	/* Create and register DMA param */
	ret = doca_argp_param_create(&dma_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to create ARGP param: %s", doca_get_error_string(ret));
	doca_argp_param_set_short_name(dma_param, "d");
	doca_argp_param_set_long_name(dma_param, "dma");
	doca_argp_param_set_description(dma_param, "DMA device name");
	doca_argp_param_set_callback(dma_param, dma_callback);
	doca_argp_param_set_type(dma_param, DOCA_ARGP_TYPE_STRING);
	doca_argp_param_set_mandatory(dma_param);
	ret = doca_argp_register_param(dma_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to register program param: %s", doca_get_error_string(ret));

	/* Create and register system OS map param */
	ret = doca_argp_param_create(&os_syms_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to create ARGP param: %s", doca_get_error_string(ret));
	doca_argp_param_set_short_name(os_syms_param, "o");
	doca_argp_param_set_long_name(os_syms_param, "osym");
	doca_argp_param_set_arguments(os_syms_param, "<path>");
	doca_argp_param_set_description(os_syms_param, "System OS symbol map path");
	doca_argp_param_set_callback(os_syms_param, os_syms_callback);
	doca_argp_param_set_type(os_syms_param, DOCA_ARGP_TYPE_STRING);
	doca_argp_param_set_mandatory(os_syms_param);
	ret = doca_argp_register_param(os_syms_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to register program param: %s", doca_get_error_string(ret));

	/* Create and register system OS type param */
	ret = doca_argp_param_create(&os_type_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to create ARGP param: %s", doca_get_error_string(ret));
	doca_argp_param_set_short_name(os_type_param, "s");
	doca_argp_param_set_long_name(os_type_param, "osty");
	doca_argp_param_set_arguments(os_type_param, "<windows|linux>");
	doca_argp_param_set_description(os_type_param, "System OS type - windows/linux");
	doca_argp_param_set_callback(os_type_param, os_type_callback);
	doca_argp_param_set_type(os_type_param, DOCA_ARGP_TYPE_STRING);
	doca_argp_param_set_mandatory(os_type_param);
	ret = doca_argp_register_param(os_type_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to register program param: %s", doca_get_error_string(ret));

	/* Create and register time interval param */
	ret = doca_argp_param_create(&time_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to create ARGP param: %s", doca_get_error_string(ret));
	doca_argp_param_set_short_name(time_param, "t");
	doca_argp_param_set_long_name(time_param, "time");
	doca_argp_param_set_arguments(time_param, "<seconds>");
	doca_argp_param_set_description(time_param, "Scan time interval in seconds");
	doca_argp_param_set_callback(time_param, time_callback);
	doca_argp_param_set_type(time_param, DOCA_ARGP_TYPE_INT);
	doca_argp_param_set_mandatory(time_param);
	ret = doca_argp_register_param(time_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to register program param: %s", doca_get_error_string(ret));

	ret = doca_argp_register_version_callback(sdk_version_callback);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to register version callback: %s", doca_get_error_string(ret));
}

static struct doca_apsh_ctx *
apsh_ctx_init(struct apsh_config *conf, struct apsh_resources *resources)
{
	int res;

	/* Create a new apsh context */
	struct doca_apsh_ctx *apsh_ctx = doca_apsh_create();

	if (apsh_ctx == NULL)
		APP_EXIT("Create lib APSH context failed");

	/* Get dma device */
	res = open_doca_device_with_property(DOCA_DEVINFO_PROPERTY_IBDEV_NAME, (uint8_t *)conf->dma_dev_name,
					     strlen(conf->dma_dev_name), &resources->dma_device, NULL);
	if (res != DOCA_SUCCESS)
		APP_EXIT("Failed to open dma device");

	/* Start apsh context */
	/* set the DMA device */
	res = doca_apsh_dma_dev_set(apsh_ctx, resources->dma_device);
	if (res != DOCA_SUCCESS) {
		app_shield_agent_cleanup(resources);
		APP_EXIT("Set dma device failed");
	}

	/* Start apsh handler and init connection to devices */
	res = doca_apsh_start(apsh_ctx);
	if (res != DOCA_SUCCESS) {
		app_shield_agent_cleanup(resources);
		APP_EXIT("Start APSH failed");
	}

	/* return value */
	return apsh_ctx;
}

static struct doca_apsh_system *
apsh_system_init(struct apsh_config *conf, struct apsh_resources *resources)
{
	int res;
	struct doca_apsh_system *apsh_system;

	res = open_remote_doca_device_with_property(resources->dma_device, DOCA_DEV_REMOTE_FILTER_NET,
						    DOCA_DEVINFO_REMOTE_PROPERTY_VUID, (uint8_t *)conf->system_vuid,
						    strlen(conf->system_vuid), &resources->system_device);
	if (res != DOCA_SUCCESS) {
		app_shield_agent_cleanup(resources);
		APP_EXIT("Failed to open remote device");
	}

	/* Create a new system handler to introspect */
	apsh_system = doca_apsh_system_create(resources->ctx);
	if (apsh_system == NULL) {
		app_shield_agent_cleanup(resources);
		APP_EXIT("Create system context failed");
	}

	/* Start system context - bare-metal */
	/* Set the system os symbol map */
	res = doca_apsh_sys_os_symbol_map_set(apsh_system, conf->system_os_symbol_map_path);
	if (res != DOCA_SUCCESS) {
		app_shield_agent_cleanup(resources);
		APP_EXIT("Set os symbols map failed");
	}

	/* Set the system memory region the apsh handler is allowed to access */
	res = doca_apsh_sys_mem_region_set(apsh_system, conf->system_mem_region_path);
	if (res != DOCA_SUCCESS) {
		app_shield_agent_cleanup(resources);
		APP_EXIT("Set mem regions map failed");
	}

	/* Set the system device for the apsh handler to use */
	res = doca_apsh_sys_dev_set(apsh_system, resources->system_device);
	if (res != DOCA_SUCCESS) {
		app_shield_agent_cleanup(resources);
		APP_EXIT("Set system device failed");
	}

	/* Set the system os type - linux/widows */
	res = doca_apsh_sys_os_type_set(apsh_system, conf->os_type);
	if (res != DOCA_SUCCESS) {
		app_shield_agent_cleanup(resources);
		APP_EXIT("Set system os type failed");
	}

	/* Start system handler and init connection to the system and the devices. */
	res = doca_apsh_system_start(apsh_system);
	if (res != DOCA_SUCCESS) {
		app_shield_agent_cleanup(resources);
		APP_EXIT("Start system failed");
	}

	/* return value */
	return apsh_system;
}

void
app_shield_agent_init(struct apsh_config *conf, struct apsh_resources *resources)
{
	/* Init basic apsh handlers */
	memset(resources, 0, sizeof(*resources));
	resources->ctx = apsh_ctx_init(conf, resources);
	resources->sys = apsh_system_init(conf, resources);
}

void
app_shield_agent_cleanup(struct apsh_resources *resources)
{
	/* free the system handler and disconnect from the devices */
	if (resources->sys != NULL)
		doca_apsh_system_destroy(resources->sys);

	/* free the apsh handler and disconnect from the devices */
	if (resources->ctx != NULL)
		doca_apsh_destroy(resources->ctx);

	/* Close the devices */
	if (resources->dma_device != NULL)
		doca_dev_close(resources->dma_device);
	if (resources->system_device != NULL)
		doca_dev_remote_close(resources->system_device);
}

struct doca_apsh_process *
get_process_by_pid(struct doca_apsh_process ***pslist, struct apsh_resources *resources, struct apsh_config *apsh_conf)
{
	struct doca_apsh_process **processes;
	int proc_count, res;
	size_t process_idx;
	typeof(apsh_conf->pid) cur_proc_pid = 0;
	struct doca_apsh_process *process = NULL;

	/* Create list of processes on remote system */
	res = doca_apsh_processes_get(resources->sys, &processes, &proc_count);
	if (res == 0)
		*pslist = processes;
	else {
		app_shield_agent_cleanup(resources);
		APP_EXIT("Get processes failed");
	}

	/* Search for the process 'pid' */
	for (process_idx = 0; process_idx < proc_count; process_idx++) {
		cur_proc_pid = doca_apsh_proc_info_get(processes[process_idx], DOCA_APSH_PROCESS_PID);

		if (apsh_conf->pid == cur_proc_pid) {
			process = processes[process_idx];
			break;
		}
	}
	if (process == NULL) {
		doca_apsh_processes_free(processes);
		app_shield_agent_cleanup(resources);
		APP_EXIT("Process (%d) was not found", apsh_conf->pid);
	}

	*pslist = processes;
	return process;
}

static doca_error_t
telemetry_register_attest_event(struct doca_telemetry_schema *schema, doca_telemetry_type_index_t *type_index)
{
	doca_error_t ret;
	/* Event type for schema. Should be consistent with event struct. */
	struct doca_telemetry_type *type;
	struct doca_telemetry_field *field;
	const int nb_fields = 5;
	int idx = 0;
	struct {
		const char *name;
		const char *desc;
		const char *type_name;
		uint16_t len;
	} fields_info[] = {
		{"timestamp", "Event timestamp", DOCA_TELEMETRY_FIELD_TYPE_TIMESTAMP, 1},
		{"pid", "Pid", DOCA_TELEMETRY_FIELD_TYPE_INT32, 1},
		{"result", "Result", DOCA_TELEMETRY_FIELD_TYPE_INT32, 1},
		{ "scan_count", "Scan Count", DOCA_TELEMETRY_FIELD_TYPE_UINT64, 1},
		{ "path", "Path", DOCA_TELEMETRY_FIELD_TYPE_CHAR, MAX_PATH},
	};

	ret = doca_telemetry_type_create(&type);
	if (ret != DOCA_SUCCESS)
		return ret;

	for (idx = 0; idx < nb_fields; idx++) {
		ret = doca_telemetry_field_create(&field);
		if (ret != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Failed to create field");
			doca_telemetry_type_destroy(type);
			return ret;
		}
		doca_telemetry_field_set_name(field, fields_info[idx].name);
		doca_telemetry_field_set_description(field, fields_info[idx].desc);
		doca_telemetry_field_set_type_name(field, fields_info[idx].type_name);
		doca_telemetry_field_set_array_length(field, fields_info[idx].len);

		ret = doca_telemetry_type_add_field(type, field);
		if (ret != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Failed to add field to type");
			doca_telemetry_field_destroy(field);
			doca_telemetry_type_destroy(type);
			return ret;
		}
	}

	/* Register type */
	ret = doca_telemetry_schema_add_type(schema, "attestation_event", type, type_index);
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to add type to schema");
		doca_telemetry_type_destroy(type);
	}
	return ret;
}

int
telemetry_start(struct doca_telemetry_schema **telemetry_schema, struct doca_telemetry_source **telemetry_source,
											struct event_indexes *indices)
{
	doca_error_t ret;
	struct doca_telemetry_schema *schema = NULL;
	struct doca_telemetry_source *source = NULL;
	char source_id_buf[MAX_HOSTNAME_LEN + 1], source_tag_buf[MAX_HOSTNAME_LEN + strlen("_tag") + 1];

	/* Creating telemetry schema */
	ret = doca_telemetry_schema_init("app_shield_agent_telemetry", &schema);
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to init the doca telemetry schema");
		return -1;
	}

	/* Register all currently supported events */
	ret = telemetry_register_attest_event(schema, &indices->attest_index);
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to register attestation event in the telemetry schema");
		goto schema_error;
	}

	/* Enable file write during the app development.
	 * Check written files under data root to make sure that data format is correct.
	 * Default max_file_size is 1 Mb, default max_file_age is 1 hour.
	 */
	doca_telemetry_schema_set_file_write_enabled(schema);
	doca_telemetry_schema_set_file_write_max_size(schema, 1 * 1024 * 1024);
	doca_telemetry_schema_set_file_write_max_age(schema, 60 * 60 * 1000000L);

	/* Activate the schema */
	ret = doca_telemetry_schema_start(schema);
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to start the doca telemetry schema");
		goto schema_error;
	}

	/* Open a telemetry connection with custom source id and tag */
	ret = doca_telemetry_source_create(schema, &source);
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create a source end point to the telemetry");
		ret = -1;
		goto schema_error;
	}

	/* Creating a unique tag and id per host */
	if (gethostname(source_id_buf, sizeof(source_id_buf)) < 0)  {
		DOCA_LOG_ERR("Gethostname failed, can't create a unique source tag and id");
		ret = -1;
		goto source_error;
	}

	strlcpy(source_tag_buf, source_id_buf, sizeof(source_tag_buf));
	strlcat(source_tag_buf, "_tag", sizeof(source_tag_buf));
	doca_telemetry_source_set_id(source, source_id_buf);
	doca_telemetry_source_set_tag(source, source_tag_buf);

	/* Initiate the DOCA telemetry source */
	ret = doca_telemetry_source_start(source);
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to establish a source connection to the telemetry");
		goto source_error;
	}

	/* Success init, return handlers */
	*telemetry_schema = schema;
	*telemetry_source = source;
	return 0;

source_error:
	doca_telemetry_source_destroy(source);
schema_error:
	doca_telemetry_schema_destroy(schema);
	return -1;
}

void
telemetry_destroy(struct doca_telemetry_schema *telemetry_schema, struct doca_telemetry_source *telemetry_source)
{
	doca_telemetry_source_destroy(telemetry_source);
	doca_telemetry_schema_destroy(telemetry_schema);
}
