/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <bsd/string.h>

#include <doca_argp.h>
#include <doca_log.h>

#include <utils.h>

#include "file_scan_core.h"

DOCA_LOG_REGISTER(FILE_SCAN::Core);

static void
calculate_mempool_size(struct file_scan_config *app_cfg)
{
	long job_size = app_cfg->chunk_size > 0 ? app_cfg->chunk_size : app_cfg->data_buffer_len;

	app_cfg->mempool_size = MAX_MATCHES_PER_JOB * ((job_size / BF2_REGEX_JOB_LIMIT) + 1);
}

void
file_scan_init(struct file_scan_config *app_cfg)
{
	doca_error_t res;
	int ret;

	/* This apps uses 1 QP */
	app_cfg->nb_qps = 1;

	calculate_mempool_size(app_cfg);

	/* find doca_dev */
	res = open_doca_device_with_property(DOCA_DEVINFO_PROPERTY_PCI_ADDR, (uint8_t *)&app_cfg->pci_address,
			sizeof(app_cfg->pci_address), &app_cfg->dev, NULL);
	if (res != DOCA_SUCCESS)
		APP_EXIT("No device matching PCI address found. Reason: %s", doca_get_error_string(res));

	/* Create a doca reg instance. */
	app_cfg->doca_regex = doca_regex_create();
	if (app_cfg->doca_regex == NULL)
		APP_EXIT("Unable to create RegEx device.");

	/* Set the regex device as the main HW accelerator. */
	ret = doca_regex_dev_add(app_cfg->doca_regex, app_cfg->dev);
	if (ret < 0)
		APP_EXIT("Unable to set RegEx device. [%s]", strerror(abs(ret)));

	/* Configure number of QPs to use. */
	ret = doca_regex_num_qps_set(app_cfg->doca_regex, app_cfg->nb_qps);
	if (ret < 0)
		APP_EXIT("Unable to configure %d QPs. [%s]", app_cfg->nb_qps, strerror(abs(ret)));

	/* Size QP-0 memory pool, this app uses only 1 QP. */
	res = doca_regex_property_matches_memory_pool_size_set(app_cfg->doca_regex, app_cfg->mempool_size);
	if (res != DOCA_SUCCESS)
		APP_EXIT("Unable to set matches mempool size. Reason: %s", doca_get_error_string(res));

	/* Set Overlap in bytes for huge job(s). */
	res = doca_regex_property_huge_job_emulation_overlap_set(app_cfg->doca_regex, app_cfg->nb_overlap_bytes);
	if (res != DOCA_SUCCESS)
		APP_EXIT("Unable to set huge job emulation overlap size. Reason: %s", doca_get_error_string(res));

	/* Attach rules file to DOCA regex */
	res = doca_regex_property_hardware_binary_rules_set(app_cfg->doca_regex, (uint8_t *)app_cfg->rules_buffer,
							    app_cfg->rules_buffer_len);
	if (res != DOCA_SUCCESS)
		APP_EXIT("Unable to program rules file. Reason: %s", doca_get_error_string(res));

	/* Start doca RegEx */
	ret = doca_regex_start(app_cfg->doca_regex);
	if (ret < 0)
		APP_EXIT("Unable to start doca RegEx. [%s]", strerror(abs(ret)));

	res = doca_buf_inventory_create("inv", 1, DOCA_BUF_EXTENSION_NONE, &app_cfg->buf_inventory);
	if (res != DOCA_SUCCESS)
		APP_EXIT("Unable to create doca_buf_inventory. Reason: %s", doca_get_error_string(res));

	res = doca_buf_inventory_start(app_cfg->buf_inventory);
	if (res != DOCA_SUCCESS)
		APP_EXIT("Unable to start doca_buf_inventory. Reason: %s", doca_get_error_string(res));

	res = doca_mmap_create("mmap", &app_cfg->mmap);
	if (res != DOCA_SUCCESS)
		APP_EXIT("Unable to create doca_mmap. Reason: %s", doca_get_error_string(res));

	res = doca_mmap_start(app_cfg->mmap);
	if (res != DOCA_SUCCESS)
		APP_EXIT("Unable to start doca_mmap. Reason: %s", doca_get_error_string(res));

	res = doca_mmap_dev_add(app_cfg->mmap, app_cfg->dev);
	if (res != DOCA_SUCCESS)
		APP_EXIT("Unable to add device to doca_mmap. Reason: %s", doca_get_error_string(res));

	res = doca_mmap_populate(app_cfg->mmap, app_cfg->data_buffer, app_cfg->data_buffer_len, 4096, NULL, NULL);
	if (res != DOCA_SUCCESS)
		APP_EXIT("Unable to register memory with doca_mmap. Reason: %s", doca_get_error_string(res));
}

void
file_scan_cleanup(struct file_scan_config *app_cfg)
{
	doca_regex_stop(app_cfg->doca_regex);
	doca_regex_destroy(app_cfg->doca_regex);
	app_cfg->doca_regex = NULL;
	if (app_cfg->mmap != NULL) {
		doca_mmap_dev_rm(app_cfg->mmap, app_cfg->dev);
		doca_mmap_stop(app_cfg->mmap);
		doca_mmap_destroy(app_cfg->mmap);
		app_cfg->mmap = NULL;
	}
	if (app_cfg->buf_inventory != NULL) {
		doca_buf_inventory_stop(app_cfg->buf_inventory);
		doca_buf_inventory_destroy(app_cfg->buf_inventory);
		app_cfg->buf_inventory = NULL;
	}
	if (app_cfg->data_buffer != NULL) {
		free(app_cfg->data_buffer);
		app_cfg->data_buffer = NULL;
	}
	if (app_cfg->rules_buffer != NULL) {
		free(app_cfg->rules_buffer);
		app_cfg->rules_buffer = NULL;
	}
	if (app_cfg->csv_fp != NULL) {
		fclose(app_cfg->csv_fp);
		app_cfg->csv_fp = NULL;
	}
	if (app_cfg->dev != NULL) {
		doca_dev_close(app_cfg->dev);
		app_cfg->dev = NULL;
	}
}

static doca_error_t
rules_callback(void *config, void *param)
{
	struct file_scan_config *app_cfg = (struct file_scan_config *)config;
	char *data_path = (char *)param;

	/* Read data file into the data buffer */
	return read_file(data_path, &app_cfg->rules_buffer, &app_cfg->rules_buffer_len);
}

static doca_error_t
data_callback(void *config, void *param)
{
	struct file_scan_config *app_cfg = (struct file_scan_config *)config;
	char *data_path = (char *)param;

	/* Read data file into the data buffer */
	return read_file(data_path, &app_cfg->data_buffer, &app_cfg->data_buffer_len);
}

static doca_error_t
pci_address_callback(void *config, void *param)
{
	struct file_scan_config *app_cfg = (struct file_scan_config *)config;
	char *pci_address = (char *)param;

	if (parse_pci_addr(pci_address, &app_cfg->pci_address) != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Invalid PCI address: \"%s\"", pci_address);
		return DOCA_ERROR_INVALID_VALUE;
	}
	return DOCA_SUCCESS;
}

static doca_error_t
csv_callback(void *config, void *param)
{
	struct file_scan_config *app_cfg = (struct file_scan_config *)config;
	char *csv_path = (char *)param;

	if (csv_path == NULL)
		return DOCA_SUCCESS;
	app_cfg->csv_fp = fopen(csv_path, "w");
	if (app_cfg->csv_fp == NULL) {
		DOCA_LOG_ERR("Failed to create CSV file. Skipping");
		return DOCA_SUCCESS;
	}
	fprintf(app_cfg->csv_fp, "Line number,Match Index,Match Length,Rule Id\n");
	return DOCA_SUCCESS;
}

static doca_error_t
chunk_callback(void *config, void *param)
{
	struct file_scan_config *app_cfg = (struct file_scan_config *)config;
	int chunk = *(int *)param;

	if (chunk < 0) {
		DOCA_LOG_ERR("Chunk size must be > 0.");
		return DOCA_ERROR_INVALID_VALUE;
	}
	app_cfg->chunk_size = (uint32_t)chunk;
	return DOCA_SUCCESS;
}

static doca_error_t
overlap_callback(void *config, void *param)
{
	struct file_scan_config *app_cfg = (struct file_scan_config *)config;
	int overlap = *(int *)param;

	if (overlap < 0) {
		DOCA_LOG_ERR("Overlap size must be > 0.");
		return DOCA_ERROR_INVALID_VALUE;
	}
	app_cfg->nb_overlap_bytes = (uint32_t)overlap;
	return DOCA_SUCCESS;
}

void
register_file_scan_params(void)
{
	doca_error_t ret;
	struct doca_argp_param *rules_param, *data_param, *pci_param, *csv_param, *chunk_param, *overlap_param;

	/* Create and register regex rules param */
	ret = doca_argp_param_create(&rules_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to create ARGP param: %s", doca_get_error_string(ret));
	doca_argp_param_set_short_name(rules_param, "r");
	doca_argp_param_set_long_name(rules_param, "rules");
	doca_argp_param_set_arguments(rules_param, "<path>");
	doca_argp_param_set_description(rules_param, "Path to compiled rules file (rof2.binary)");
	doca_argp_param_set_callback(rules_param, rules_callback);
	doca_argp_param_set_type(rules_param, DOCA_ARGP_TYPE_STRING);
	doca_argp_param_set_mandatory(rules_param);
	ret = doca_argp_register_param(rules_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to register program param: %s", doca_get_error_string(ret));

	/* Create and register data to scan param */
	ret = doca_argp_param_create(&data_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to create ARGP param: %s", doca_get_error_string(ret));
	doca_argp_param_set_short_name(data_param, "d");
	doca_argp_param_set_long_name(data_param, "data");
	doca_argp_param_set_arguments(data_param, "<path>");
	doca_argp_param_set_description(data_param, "Path to data file");
	doca_argp_param_set_callback(data_param, data_callback);
	doca_argp_param_set_type(data_param, DOCA_ARGP_TYPE_STRING);
	doca_argp_param_set_mandatory(data_param);
	ret = doca_argp_register_param(data_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to register program param: %s", doca_get_error_string(ret));

	/* Create and register regex pci address param */
	ret = doca_argp_param_create(&pci_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to create ARGP param: %s", doca_get_error_string(ret));
	doca_argp_param_set_short_name(pci_param, "p");
	doca_argp_param_set_long_name(pci_param, "pci-addr");
	doca_argp_param_set_arguments(pci_param, "<address>");
	doca_argp_param_set_description(pci_param, "Set PCI address of the RXP engine to use");
	doca_argp_param_set_callback(pci_param, pci_address_callback);
	doca_argp_param_set_type(pci_param, DOCA_ARGP_TYPE_STRING);
	doca_argp_param_set_mandatory(pci_param);
	ret = doca_argp_register_param(pci_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to register program param: %s", doca_get_error_string(ret));

	/* Create and register regex pci address param */
	ret = doca_argp_param_create(&csv_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to create ARGP param: %s", doca_get_error_string(ret));
	doca_argp_param_set_short_name(csv_param, "o");
	doca_argp_param_set_long_name(csv_param, "output-csv");
	doca_argp_param_set_arguments(csv_param, "<path>");
	doca_argp_param_set_description(csv_param, "Path to the output of the CSV file");
	doca_argp_param_set_callback(csv_param, csv_callback);
	doca_argp_param_set_type(csv_param, DOCA_ARGP_TYPE_STRING);
	ret = doca_argp_register_param(csv_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to register program param: %s", doca_get_error_string(ret));

	/* Create and register chunk size param */
	ret = doca_argp_param_create(&chunk_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to create ARGP param: %s", doca_get_error_string(ret));
	doca_argp_param_set_short_name(chunk_param, "c");
	doca_argp_param_set_long_name(chunk_param, "chunk-size");
	doca_argp_param_set_arguments(chunk_param, "<bytes_number>");
	doca_argp_param_set_description(chunk_param,
					"Chunk size of each job sent to the regex,  use 0 to send the file as 1 chunk");
	doca_argp_param_set_callback(chunk_param, chunk_callback);
	doca_argp_param_set_type(chunk_param, DOCA_ARGP_TYPE_INT);
	ret = doca_argp_register_param(chunk_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to register program param: %s", doca_get_error_string(ret));

	/* Create and register regex pci address param */
	ret = doca_argp_param_create(&overlap_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to create ARGP param: %s", doca_get_error_string(ret));
	doca_argp_param_set_short_name(overlap_param, "s");
	doca_argp_param_set_long_name(overlap_param, "overlap-size");
	doca_argp_param_set_arguments(overlap_param, "<bytes_number>");
	doca_argp_param_set_description(overlap_param, "Number of bytes of overlap to use for huge jobs");
	doca_argp_param_set_callback(overlap_param, overlap_callback);
	doca_argp_param_set_type(overlap_param, DOCA_ARGP_TYPE_INT);
	ret = doca_argp_register_param(overlap_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to register program param: %s", doca_get_error_string(ret));

	/* Register version callback for DOCA SDK & RUNTIME */
	ret = doca_argp_register_version_callback(sdk_version_callback);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to register version callback: %s", doca_get_error_string(ret));
}
