/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <string.h>

#include <doca_argp.h>

#include <utils.h>

#include "secure_channel_core.h"

#define MAX_MSG 10
#define SERVER_NAME "secure_channel_server"

DOCA_LOG_REGISTER(SECURE_CHANNEL::Core);

doca_error_t
secure_channel_client(struct doca_comm_channel_ep_t *ep, struct doca_comm_channel_addr_t **peer_addr,
		      struct secure_channel_config *app_cfg)
{
	char rcv_buf[MAX_MSG_SIZE];
	int client_msg_len = strlen(app_cfg->client_msg) + 1;
	int msg_count;
	size_t msg_len;
	doca_error_t ret;

	ret = doca_comm_channel_ep_connect(ep, SERVER_NAME, peer_addr);
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Couldn't establish a connection with the server node: %s", doca_get_error_string(ret));
		return ret;
	}

	DOCA_LOG_INFO("Connection to server was established successfully");

	for (msg_count = 0; msg_count < app_cfg->nb_send_msg; msg_count++) {

		DOCA_LOG_DBG("Sending message to server: %s", app_cfg->client_msg);
		ret = doca_comm_channel_ep_sendto(ep, app_cfg->client_msg, client_msg_len, 0,
						  *peer_addr);
		if (ret != DOCA_SUCCESS) {
			DOCA_LOG_WARN("Message number %d was not sent: %s", msg_count,
				      doca_get_error_string(ret));
			continue;
		}
		msg_len = MAX_MSG_SIZE;
		ret = doca_comm_channel_ep_recvfrom(ep, (void *)rcv_buf, &msg_len, 0,
						    peer_addr);
		if (ret != DOCA_SUCCESS) {
			DOCA_LOG_WARN("Message was not received: %s", doca_get_error_string(ret));
			continue;
		}
		rcv_buf[MAX_MSG_SIZE - 1] = '\0';
		DOCA_LOG_INFO("Server response: %s", rcv_buf);
	}

	return DOCA_SUCCESS;
}

doca_error_t
secure_channel_server(struct doca_comm_channel_ep_t *ep, struct doca_comm_channel_addr_t **peer_addr)
{
	doca_error_t ret;
	size_t msg_len;
	char rcv_buf[MAX_MSG_SIZE];

	ret = doca_comm_channel_ep_listen(ep, SERVER_NAME);
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Secure Channel server couldn't start listening: %s", doca_get_error_string(ret));
		return  ret;
	}

	DOCA_LOG_INFO("Server started Listening, waiting for new connections");

	while (1) {
		msg_len = MAX_MSG_SIZE;
		ret = doca_comm_channel_ep_recvfrom(ep, (void *)rcv_buf, &msg_len, 0,
						    peer_addr);
		if (ret != DOCA_SUCCESS) {
			DOCA_LOG_WARN("Message was not received: %s", doca_get_error_string(ret));
			continue;
		}

		rcv_buf[MAX_MSG_SIZE - 1] = '\0';
		DOCA_LOG_INFO("Received message: %s", rcv_buf);

		size_t nb_bytes = msg_len;

		ret = doca_comm_channel_ep_sendto(ep, rcv_buf, nb_bytes, 0, *peer_addr);
		if (ret != DOCA_SUCCESS)
			DOCA_LOG_WARN("Response was not sent successfully: %s", doca_get_error_string(ret));
	}
}

static doca_error_t
ep_mode_callback(void *config, void *param)
{
	struct secure_channel_config *app_cfg = (struct secure_channel_config *)config;
	const char *mode = (char *)param;

	if (strcmp(mode, "client") == 0)
		app_cfg->mode = SC_EP_MODE_CLIENT;
	else if (strcmp(mode, "server") == 0)
		app_cfg->mode = SC_EP_MODE_SERVER;
	else {
		DOCA_LOG_ERR("Illegal end-point mode = [%s]", mode);
		return DOCA_ERROR_INVALID_VALUE;
	}

	return DOCA_SUCCESS;
}

static doca_error_t
messages_amount_callback(void *config, void *param)
{
	struct secure_channel_config *app_cfg = (struct secure_channel_config *)config;
	int nb_send_msg = *(int *)param;

	if (nb_send_msg < 1) {
		DOCA_LOG_ERR("Amount of messages to be sent by the client is less than 1");
		return DOCA_ERROR_INVALID_VALUE;
	}

	app_cfg->nb_send_msg = nb_send_msg;
	return DOCA_SUCCESS;
}

static doca_error_t
message_callback(void *config, void *param)
{
	struct secure_channel_config *app_cfg = (struct secure_channel_config *)config;
	char *rcv_send_msg = (char *)param;
	int send_msg_len = strnlen(rcv_send_msg, MAX_MSG_SIZE);

	memcpy(app_cfg->client_msg, rcv_send_msg, send_msg_len);

	if (send_msg_len == MAX_MSG_SIZE) {
		DOCA_LOG_WARN("Message exceeded buffer size, trimming it to maximum size");
		app_cfg->client_msg[MAX_MSG_SIZE - 1] = '\0';
	}
	return DOCA_SUCCESS;
}

static doca_error_t
args_validation_callback(void *cfg, void *value)
{
	struct secure_channel_config *app_cfg = (struct secure_channel_config *)cfg;

	if (app_cfg->mode == SC_EP_MODE_SERVER && ((app_cfg->client_msg[0] != '\0') || (app_cfg->nb_send_msg != 0))) {
		DOCA_LOG_ERR(
			"Secure Channel server can not be configured with client message nor number of messages to be send");
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (app_cfg->mode == SC_EP_MODE_CLIENT && ((app_cfg->client_msg[0] == '\0') || (app_cfg->nb_send_msg == 0))) {
		DOCA_LOG_ERR(
			"Secure Channel client can not be configured without a message and the amount of messages to be send");
		return DOCA_ERROR_INVALID_VALUE;
	}
	return DOCA_SUCCESS;
}

void
register_secure_channel_params()
{
	doca_error_t ret;
	struct doca_argp_param *ep_mode_param, *message_param, *messages_amount_param;

	/* Create and register server param */
	ret = doca_argp_param_create(&ep_mode_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to create ARGP param: %s", doca_get_error_string(ret));
	doca_argp_param_set_short_name(ep_mode_param, "em");
	doca_argp_param_set_long_name(ep_mode_param, "ep-mode");
	doca_argp_param_set_description(ep_mode_param, "Set end-point mode {server, client}");
	doca_argp_param_set_callback(ep_mode_param, ep_mode_callback);
	doca_argp_param_set_type(ep_mode_param, DOCA_ARGP_TYPE_STRING);
	doca_argp_param_set_mandatory(ep_mode_param);
	ret = doca_argp_register_param(ep_mode_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to register program param: %s", doca_get_error_string(ret));

	/* Create and register message to send param */
	ret = doca_argp_param_create(&message_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to create ARGP param: %s", doca_get_error_string(ret));
	doca_argp_param_set_short_name(message_param, "m");
	doca_argp_param_set_long_name(message_param, "message");
	doca_argp_param_set_description(message_param, "Message to send by the client");
	doca_argp_param_set_callback(message_param, message_callback);
	doca_argp_param_set_type(message_param, DOCA_ARGP_TYPE_STRING);
	ret = doca_argp_register_param(message_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to register program param: %s", doca_get_error_string(ret));

	/* Create and register number of message param */
	ret = doca_argp_param_create(&messages_amount_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to create ARGP param: %s", doca_get_error_string(ret));
	doca_argp_param_set_short_name(messages_amount_param, "n");
	doca_argp_param_set_long_name(messages_amount_param, "num-msgs");
	doca_argp_param_set_description(messages_amount_param, "Number of messages to send by the client");
	doca_argp_param_set_callback(messages_amount_param, messages_amount_callback);
	doca_argp_param_set_type(messages_amount_param, DOCA_ARGP_TYPE_INT);
	ret = doca_argp_register_param(messages_amount_param);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to register program param: %s", doca_get_error_string(ret));

	/* Register version callback for DOCA SDK & RUNTIME */
	ret = doca_argp_register_version_callback(sdk_version_callback);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to register version callback: %s", doca_get_error_string(ret));

	/* Register application callback */
	ret = doca_argp_register_validation_callback(args_validation_callback);
	if (ret != DOCA_SUCCESS)
		APP_EXIT("Failed to register program validation callback: %s", doca_get_error_string(ret));
}

void
configure_ep_qp_attributes(struct doca_comm_channel_init_attr *attr)
{
	attr->msgsize = MAX_MSG_SIZE;
	attr->maxmsgs = MAX_MSG;
	attr->cookie = 0;
	attr->flags = 0;
}

doca_error_t
create_secure_channel_ep(struct doca_comm_channel_ep_t **ep,
			 struct doca_comm_channel_init_attr *attr)
{
	doca_error_t ret;

	ret = doca_comm_channel_ep_create(attr, ep);

	if (ret != DOCA_SUCCESS)
		DOCA_LOG_ERR("Failed to create Secure Channel endpoint: %s", doca_get_error_string(ret));

	return ret;
}
